#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from KDEpy.NaiveKDE import NaiveKDE
from KDEpy.TreeKDE import TreeKDE
from KDEpy.FFTKDE import FFTKDE

__version__ = '0.5.6'
__author__ = 'tommyod'

TreeKDE = TreeKDE
NaiveKDE = NaiveKDE
FFTKDE = FFTKDE
